@@ -34,7 +34,7 @@
 #include "gdkwayland.h"
 #include <gdk-pixbuf/gdk-pixbuf.h>
 
-#include <wayland-cursor.h>
+#include <cursor/wayland-cursor.h>
 
 #define GDK_TYPE_WAYLAND_CURSOR              (_gdk_wayland_cursor_get_type ())
 #define GDK_WAYLAND_CURSOR(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), GDK_TYPE_WAYLAND_CURSOR, GdkWaylandCursor))
@@ -149,9 +149,8 @@ _gdk_wayland_cursor_update (GdkWaylandDisplay *display_wayland,
   if (cursor->name == NULL)
     return FALSE;
 
-  theme = _gdk_wayland_display_get_scaled_cursor_theme (display_wayland,
-                                                        cursor->scale);
-  c = wl_cursor_theme_get_cursor (theme, cursor->name);
+  theme = _gdk_wayland_display_get_cursor_theme (display_wayland);
+  c = wl_cursor_theme_get_cursor (theme, cursor->name, cursor->scale);
   if (!c)
     {
       const char *fallback;
@@ -159,15 +158,15 @@ _gdk_wayland_cursor_update (GdkWaylandDisplay *display_wayland,
       fallback = name_fallback (cursor->name);
       if (fallback)
         {
-          c = wl_cursor_theme_get_cursor (theme, name_fallback (cursor->name));
+          c = wl_cursor_theme_get_cursor (theme, fallback, cursor->scale);
           if (!c)
-            c = wl_cursor_theme_get_cursor (theme, "left_ptr");
+            c = wl_cursor_theme_get_cursor (theme, "left_ptr", cursor->scale);
         }
     }
 
   if (!c)
     {
-      g_warning (G_STRLOC ": Unable to load %s from the cursor theme", cursor->name);
+      g_message ("Unable to load %s from the cursor theme", cursor->name);
       return FALSE;
     }
 
@@ -259,6 +258,14 @@ _gdk_wayland_cursor_get_buffer (GdkCursor *cursor,
       if (wayland_cursor->surface.cairo_surface)
         return _gdk_wayland_shm_surface_get_wl_buffer (wayland_cursor->surface.cairo_surface);
     }
+  else
+    {
+      *hotspot_x = 0;
+      *hotspot_y = 0;
+      *w = 0;
+      *h = 0;
+      *scale = 1;
+    }
 
   return NULL;
 }
@@ -297,17 +304,15 @@ _gdk_wayland_cursor_set_scale (GdkCursor *cursor,
     GDK_WAYLAND_DISPLAY (gdk_cursor_get_display (cursor));
   GdkWaylandCursor *wayland_cursor = GDK_WAYLAND_CURSOR (cursor);
 
-  if (scale > GDK_WAYLAND_MAX_THEME_SCALE)
-    {
-      g_warning (G_STRLOC ": cursor theme size %u too large", scale);
-      scale = GDK_WAYLAND_MAX_THEME_SCALE;
-    }
-
   if (wayland_cursor->scale == scale)
     return;
 
   wayland_cursor->scale = scale;
 
+  /* Blank cursor case */
+  if (g_strcmp0 (wayland_cursor->name, "none") == 0)
+    return;
+
   _gdk_wayland_cursor_update (display_wayland, wayland_cursor);
 }
 
@@ -332,34 +337,43 @@ _gdk_wayland_display_get_cursor_for_name_with_scale (GdkDisplay  *display,
                                                      const gchar *name,
                                                      guint        scale)
 {
-  GdkWaylandCursor *private;
+  GdkWaylandCursor *wayland_cursor;
   GdkWaylandDisplay *display_wayland = GDK_WAYLAND_DISPLAY (display);
 
   g_return_val_if_fail (GDK_IS_DISPLAY (display), NULL);
 
-  private = g_hash_table_lookup (display_wayland->cursor_cache, name);
-  if (private)
-    return GDK_CURSOR (g_object_ref (private));
+  wayland_cursor = g_hash_table_lookup (display_wayland->cursor_cache, name);
+  if (wayland_cursor && wayland_cursor->scale == scale)
+    return GDK_CURSOR (g_object_ref (wayland_cursor));
 
-  private = g_object_new (GDK_TYPE_WAYLAND_CURSOR,
-                          "cursor-type", GDK_CURSOR_IS_PIXMAP,
-                          "display", display,
-                          NULL);
-  private->name = g_strdup (name);
-  private->scale = scale;
+  wayland_cursor = g_object_new (GDK_TYPE_WAYLAND_CURSOR,
+                                 "cursor-type", GDK_CURSOR_IS_PIXMAP,
+                                 "display", display,
+                                 NULL);
 
   /* Blank cursor case */
   if (!name || g_str_equal (name, "none") || g_str_equal (name, "blank_cursor"))
-    return GDK_CURSOR (private);
+    {
+      wayland_cursor->name = g_strdup ("none");
+      wayland_cursor->scale = scale;
 
-  if (!_gdk_wayland_cursor_update (display_wayland, private))
-    return GDK_CURSOR (private);
+      return GDK_CURSOR (wayland_cursor);
+    }
+
+  wayland_cursor->name = g_strdup (name);
+  wayland_cursor->scale = scale;
+
+  if (!_gdk_wayland_cursor_update (display_wayland, wayland_cursor))
+    {
+      g_object_unref (wayland_cursor);
+      return NULL;
+    }
 
   /* Insert into cache. */
-  g_hash_table_insert (display_wayland->cursor_cache,
-                       private->name,
-                       g_object_ref (private));
-  return GDK_CURSOR (private);
+  g_hash_table_replace (display_wayland->cursor_cache,
+                        wayland_cursor->name,
+                        g_object_ref (wayland_cursor));
+  return GDK_CURSOR (wayland_cursor);
 }
 
 GdkCursor *
@@ -444,6 +458,8 @@ _gdk_wayland_display_get_cursor_for_surface (GdkDisplay *display,
       cursor->surface.scale = (int)sx;
       cursor->surface.width = cairo_image_surface_get_width (surface);
       cursor->surface.height = cairo_image_surface_get_height (surface);
+      cursor->surface.hotspot_x *= sx;
+      cursor->surface.hotspot_y *= sx;
     }
   else
     {
