@@ -40,17 +40,13 @@ static GtkCssValue *
 gtk_css_value_corner_compute (GtkCssValue             *corner,
                               guint                    property_id,
                               GtkStyleProviderPrivate *provider,
-			      int                      scale,
-                              GtkCssStyle    *values,
-                              GtkCssStyle    *parent_values,
-                              GtkCssDependencies      *dependencies)
+                              GtkCssStyle             *style,
+                              GtkCssStyle             *parent_style)
 {
   GtkCssValue *x, *y;
-  GtkCssDependencies x_deps, y_deps;
 
-  x = _gtk_css_value_compute (corner->x, property_id, provider, scale, values, parent_values, &x_deps);
-  y = _gtk_css_value_compute (corner->y, property_id, provider, scale, values, parent_values, &y_deps);
-  *dependencies = _gtk_css_dependencies_union (x_deps, y_deps);
+  x = _gtk_css_value_compute (corner->x, property_id, provider, style, parent_style);
+  y = _gtk_css_value_compute (corner->y, property_id, provider, style, parent_style);
   if (x == corner->x && y == corner->y)
     {
       _gtk_css_value_unref (x);
@@ -136,7 +132,7 @@ _gtk_css_corner_value_parse (GtkCssParser *parser)
   if (x == NULL)
     return NULL;
 
-  if (!_gtk_css_parser_has_number (parser))
+  if (!gtk_css_number_value_can_parse (parser))
     y = _gtk_css_value_ref (x);
   else
     {
